/*******************************************************************************

   Termostato Diferencial

   c/ PIC 16F688

   Autor: Cludio Lrios

   Data: 29/07/2016

   Utiliza 2 Sensores NTC 10k

   Aplicao: Aquecedores solares - controlando bomba de circulao de gua

   Uso didtico

   Este arquivo  parte integrante do blog Lrios.tecnologia.ws

*******************************************************************************/
#include <16F688.h>
#device adc=10
#use delay(clock=8000000)
#fuses NOWDT,INTRC_IO, NOCPD, NOPROTECT, NOMCLR, NOPUT, NOBROWNOUT



//configurao dos ports
#byte  trisc = 0x87
#byte  trisa = 0x85
#byte  portc = 0x07
#byte  porta = 0x05
#byte  ansel = 0x91
#byte  cmcon0  = 0x19
#byte  cmcon1  = 0x1a
#bit   rp0     = 0x03.5


//bits do lcd
#bit   db4       = 0x05.0 //ra0  db4 pino 13
#bit   db5       = 0x05.1 //ra1  db5 pino 12
#bit   db6       = 0x05.2 //ra2  db6 pino 11
#bit   db7       = 0x07.0 //rc0  db7 pino 10
#bit   en        = 0x07.1 //rc1  lcd 'enable' pino 9
#bit   rs        = 0x07.2 //rc2  lcd 'rs' - comando/dado pino 8
//     rw ao gnd.

//bits
#bit   sw_setup  = 0x07.4 //rc4 setup sw
#bit   sw_inc    = 0x05.3 //ra3 inc sw
#bit   sw_dec    = 0x07.5 //rc5 dec sw
#bit   rele      = 0x05.5 //sada do rele da bomba de gua

//defaults
#define DELTA_ON_FCTR 8  //delta on (default)
#define DELTA_OFF_FCTR 4  //delta off (default)
#define TC_MAX_FCTR  85   //temperatura mxima no coletor solar (default)
#define TC_MIN_FCTR  4   //temperatura mnima no coletor solar (default)

//limites de ajustes
#define LIMITE_SUP_TC_MAX 150  //limite mximo de temperatura no coletor solar
#define LIMITE_INF_TC_MAX 50  //limite mnimo de temperatura no coletor solar
#define LIMITE_MAX_DELTA_ON  20 // limite mximo do delta on
#define LIMITE_MIN_DELTA_ON  1  //limite mnmo do delta on
#define LIMITE_INF_TC_MIN 1 //limite da mnima temperatura p/ ligar circulao
#define LIMITE_SUP_TC_MIN 5 //limite da mxima temperatura p/ ligar circulao

//endereos na eeprom
#define END_DELTA_ON 0   //posio armazenada na eeprom p/ delta on
#define END_DELTA_OFF 1  //posio armazenada na eeprom p/ delta off
#define END_TC_MAX_FCTR 2 //posio armazenada na eeprom p/ Tc_max
#define END_TC_MIN_FCTR 3 //posio armazenada na eeprom p/ Tc_min

//endereo de caracteres especiais na CGRAM
#define CGRAM_QUINA_UPLEFT    0  //caracter especial do movimento circular
#define CGRAM_QUINA_UPRIGHT   1  //caracter especial do movimento circular
#define CGRAM_QUINA_DOWNLEFT  2  //caracter especial do movimento circular
#define CGRAM_QUINA_DOWNRIGHT 3  //caracter especial do movimento circular
#define CGRAM_TRACO_A         4  //caracter especial do movimento circular
#define CGRAM_TRACO_B         5  //caracter especial do movimento circular
#DEFINE CGRAM_SETA_D          6  //caracter do simbolo 'seta cheia a direita'
#DEFINE CGRAM_SIMBOLO_DELTA   7  //caracter do simbolo 'delta'


//eeprom com valores 'default' ao programar o pic
#rom 0x2100 = {DELTA_ON_FCTR,DELTA_OFF_FCTR,TC_MAX_FCTR,TC_MIN_FCTR}

//variavis globais
float Temp,R_ntc,c,Tcol,T_reserv;//variaveis de ponto flutuante
unsigned int8 dt_on,dt_off,sup_aq,gelo;//variaveis auxiliares
unsigned int1 f_ligado,f_sup_aq,f_gelo;//flags

//constantes
float const beta = 3950.0;// valor beta do ntc (altere conforme o ntc escolhido)
float const Ro  = 10000.0;  //resistncia nominal a 25C do NTC escolhido
float const _T0 = 1.0/298.15;//inverso de 298,15 K (25c)
float const vref_adc = 5.0; //tenso de referncia do conversor AD
float const resol_adc = 1023.0 ;//resoluo do conversor AD - 1
float const res_pull_ntc = 10000.0; //resistor de pullup junto ao NTC

//==============================================================================
//                     ROTINAS PARA LCD
//==============================================================================
//clock para lcd
void lcd_clk(void){
 en=1;
 delay_us(58);
 en=0;
 }

//envia nible para lcd
 void env_nib(int8 b){

  db4=0;db5=0;db6=0;db7=0;
  rp0=1;
  db4=0;db5=0;db6=0;db7=0;
  en=0;rs=0;//rw=0;
  rp0=0;
  if(bit_test(b,7))db7=1;
  if(bit_test(b,6))db6=1;
  if(bit_test(b,5))db5=1;
  if(bit_test(b,4))db4=1;
  lcd_clk();
 }

 //envia comandos para o lcd
 void env_lcd_c(int8 val){
 rs=0;
 env_nib(val);
 env_nib(swap(val));
 delay_ms(10);
}
//envia dados individuais para o lcd
 void env_lcd_d(int8 val){
 rs=1;
 env_nib(val);
 env_nib(swap(val));
 }

 //envia string de caracteres
 void lcd_putc( char c) {
   switch (c) {
     case '\f'   : env_lcd_c(0x01);
                   delay_ms(2);
                                           break;  //apaga a tela
     case '\r'   : env_lcd_c(0x80);        break;  //linha 1
     case '\n'   : env_lcd_c(0xc0);        break;  //linha 2
     case '\b'   : env_lcd_c(0x10);        break;  //volta um digito
     default     : env_lcd_d(c);           break;  //envia dado
   }
}

//==============================================================================
//    ROTINA DE INICIALIZAO DO LCD
//==============================================================================

void lcd_init(void){
 int8 a;
  db4=0;db5=0;db6=0;db7=0;
  delay_ms(40);
  db4=1; db5=1;
  for(a=0;a<3;a++) lcd_clk();
  db4=0;
  lcd_clk();
  env_lcd_c(0x28);//ajusta para 4 bits o modo de envio
  env_lcd_c(0x06);// no shift
  env_lcd_c(0x0c);//Display ON, Cursor OFF, Blink OFF
  env_lcd_c(0x01);//limpa tela

}
/*******************************************************************************
         Rotina da Programao de Caracteres Especiais na CGRAM
*******************************************************************************/
void prog_special_char(){
   env_lcd_c(0x40);//enderea CGRAM
   env_lcd_d(0x0f); //quina upesq  0
   env_lcd_d(0x10);
   env_lcd_d(0x10);
   env_lcd_d(0x10);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   //------------------------
   env_lcd_d(0x1e); //quina updir 1
   env_lcd_d(0x01);
   env_lcd_d(0x01);
   env_lcd_d(0x01);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
 //------------------------------------
   env_lcd_d(0x00); //quina downesq 2
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x10);
   env_lcd_d(0x10);
   env_lcd_d(0x10);
   env_lcd_d(0x0f);
   env_lcd_d(0x00);
 //------------------------------------
   env_lcd_d(0x00); //quina downdir 3
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x01);
   env_lcd_d(0x01);
   env_lcd_d(0x01);
   env_lcd_d(0x0f);
   env_lcd_d(0x00);
 //-----------------------------------
   env_lcd_d(0x1c); //trao upesq downdir  4
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x07);
   env_lcd_d(0x00);
    //-----------------------------------
   env_lcd_d(0x07); //trao updir downesq 5
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x1c);
   env_lcd_d(0x00);
    //-----------------------------------

   env_lcd_d(0x10); //seta -> 6
   env_lcd_d(0x18);
   env_lcd_d(0x1c);
   env_lcd_d(0x1e);
   env_lcd_d(0x1c);
   env_lcd_d(0x18);
   env_lcd_d(0x10);
   env_lcd_d(0x00);

//----------------------------------------
   env_lcd_d(0x00); //triangulo equilatero (delta)  7
   env_lcd_d(0x00);
   env_lcd_d(0x00);
   env_lcd_d(0x04);
   env_lcd_d(0x0a);
   env_lcd_d(0x11);
   env_lcd_d(0x1f);
   env_lcd_d(0x00);
 }
/*******************************************************************************
         Rotina Matemtica p/ Logaritmo Natural -  Log_n(x)
*******************************************************************************/

float log_n(float x){

float const fpl[4] = {0.45145214, -9.0558803, 26.940971, -19.860189};
float const fql[4] = {1.0000000,  -8.1354259, 16.780517, -9.9300943};
float z, res, res1, zq;
signed int8 neg;
#define FLN2 0.6931471806

   z = x;
   if (z == 1.0) { res = 0.0; }
   else{
      *(&z) = 126;//limita a faixa entre +0,5 a -0,5
      z = (z - 1.0) / (z + 1.0);
      zq= z * z;
      res = fpl[0] * zq + fpl[1];
      res = res * zq + fpl[2];
      res = res * zq + fpl[3];
      res1 = fql[0] * zq + fql[1];
      res1 = res1 * zq + fql[2];
      res1 = res1 * zq + fql[3];
      res = z * res / res1;
      neg = * (&x) - 126;
      if (neg<0) res1 = -(float)-neg;
      else  res1 = (float) neg;
      res += res1 * FLN2;
   }
      return(res);
  }

/*******************************************************************************
         Rotina da Medio da Temperatura de Sensores
*******************************************************************************/

float get_temp( unsigned int8 canal){
    unsigned int16 res;
    float valor,R_ntc;
    unsigned int8 a;

    set_adc_channel(canal); //ajusta para canal desejado
    delay_us(20);
    res=READ_ADC();// l valor da tenso
    for(a=0;a<15;a++){
     res+=READ_ADC();
     }
     res/=16;
     //obtemos o valor da resistncia do NTC
     valor=vref_adc * res/resol_adc ;
     R_ntc=valor * res_pull_ntc / (vref_adc - valor);

     //obtemos o valor da temperatura do sensor em graus Centigrados
     //Formula de Steinhart-Hart
     Temp=log_n(R_ntc/Ro);//obtm o log n do valor R lido/resistencia a 25 graus
     Temp/= beta;//divide pelo 'beta'
     Temp+=_T0;//soma o valor nominal invertido
     Temp=1/Temp;//inverte
     Temp-=273,15;//converte para graus centigrados
     return (Temp);

}

/*******************************************************************************
         Rotinas de Escrita/Leitura na EEprom Interna
*******************************************************************************/
void escreve_eep(unsigned int8 address,unsigned int8 dado){//escreve EEprom
  write_eeprom(address,dado);
}
unsigned int8 le_eep(unsigned int8 address){//le EEprom
  return (read_eeprom(address));
}

/*******************************************************************************
         Rotinas de Delay e Debouncing
*******************************************************************************/
void delay_dbc(){ delay_ms(25);}
void dbc_sw_setup(){ while(!sw_setup)delay_dbc();}
void dbc_sw_dec(){ while(!sw_dec)delay_dbc();}
void dbc_sw_inc(){ while(!sw_inc)delay_dbc();}
/*******************************************************************************
         Rotinas de Setup:
  1) Ajuste default ou manual
  2) Ajuste do delta on
  3) Ajuste do delta off
  4) Ajuste de temperatura mxima permitida para ligar a bomba de circulao
  5) Ajuste de temperatura mnima no coletor para iniciar a circulao (gelo).
*******************************************************************************/

void testa_modo_prog(){
  unsigned int8 val,val1, passo=0;
  unsigned int16 a;
  #define TEMPO_PROG 2000 //(5 seg) tempo com tecla aperta p/ entrar em modo de setup
     if(sw_setup) return;

     for(a=0;a<TEMPO_PROG;A++){
      delay_ms(1);
      if(sw_setup)return;
     }
     env_lcd_c(1);//apaga tela
     printf(lcd_putc,"\r SetUp Mode: ");
     delay_ms(1000);//tempo de apresentao
     dbc_sw_setup();//aguarda soltar sw de setup
     env_lcd_c(1);//apaga tela

     while(passo<9){//loop da programao
//=========================================================================
//                Ajusta Default ou Manual
//=========================================================================

     if(passo==0){
     printf(lcd_putc,"\r Fctr defaults?  ");
     printf(lcd_putc,"\n NO         YES ");
     if(!sw_inc){//yes
      dbc_sw_inc();//aguarda soltar sw dec
      escreve_eep(END_DELTA_ON,DELTA_ON_FCTR);//recoloca valores 'defaults'
      escreve_eep(END_DELTA_OFF,DELTA_OFF_FCTR);
      escreve_eep(END_TC_MAX_FCTR,TC_MAX_FCTR);
      escreve_eep(END_TC_MIN_FCTR,TC_MIN_FCTR);
      env_lcd_c(1);//limpa tela
      passo=9;
     }
     if(!sw_dec){//NO
      dbc_sw_dec();//aguarda soltar boto
      passo++;
     }
      delay_dbc();
     }//if(passo=0)

     if(passo==1) {env_lcd_c(1); passo++;}//limpa tela
  //=========================================================================
  //                    Ajusta delta on
  //=========================================================================

     if(passo==2){

     printf(lcd_putc,"\r New ");
     env_lcd_d(CGRAM_SIMBOLO_DELTA);//mostra simbolo de 'delta' da CGRAM
     printf(lcd_putc,"ON Setup:");
     env_lcd_c(0xc6);//posiciona print
     printf(lcd_putc,"%u",le_eep(END_DELTA_ON));
     env_lcd_d(223);//envia  simbolo de graus
     printf(lcd_putc,"C ");//envia C
      if(!sw_inc){//testa ws dec
      dbc_sw_inc();//aguarda soltar boto
      val=le_eep(END_DELTA_ON);
      if(val<LIMITE_MAX_DELTA_ON){
       val++;
       escreve_eep(END_DELTA_ON,val);
      }
     }
     if(!sw_dec){//testa ws dec
      dbc_sw_dec();//aguarda soltar boto
      val=le_eep(END_DELTA_ON);
      if(val>LIMITE_MIN_DELTA_ON){
        val--;
        escreve_eep(END_DELTA_ON,val);
      }
     }

       if(!sw_setup){passo++; dbc_sw_setup();}
       delay_dbc();
     }//if(passo==2)

      if(passo==3) {env_lcd_c(1); passo++;}//limpa tela
 //=========================================================================
  //                    Ajusta delta off
  //=========================================================================
      if(passo==4) {
       printf(lcd_putc,"\r New ");
     env_lcd_d(CGRAM_SIMBOLO_DELTA);//mostra simbolo de 'delta' da CGRAM
     printf(lcd_putc,"OFF Setup:");
     env_lcd_c(0xc6);//posiciona print
     printf(lcd_putc,"%u",le_eep(END_DELTA_OFF));
     env_lcd_d(223);//envia  simbolo de graus
     printf(lcd_putc,"C ");//envia C

      if(!sw_inc){//testa ws dec
      dbc_sw_inc();//aguarda soltar boto
      val=le_eep(END_DELTA_OFF);
      val++;
      val1=le_eep(END_DELTA_ON);
      if(val<val1)escreve_eep(END_DELTA_OFF,val);
     }
     if(!sw_dec){//testa ws dec
      dbc_sw_dec();//aguarda soltar boto
      val=le_eep(END_DELTA_OFF);
      if(val>LIMITE_MIN_DELTA_ON){
       val--;
       escreve_eep(END_DELTA_OFF,val);
      }
     }

       if(!sw_setup){passo++; dbc_sw_setup();}
       delay_dbc();
     }//if(passo==4)


     if(passo==5) {env_lcd_c(1); passo++;}//limpa tela
  //=========================================================================
  //     Ajusta Temperatura Maxima do Coletor Solar p/ desligar Bomba
  //     obs. Proteo da tubulao contra excesso de temperatura
  //=========================================================================

    if(passo==6) {
      printf(lcd_putc,"\rNew TcMax Setup:");
      env_lcd_c(0xc6);//posiciona print
      printf(lcd_putc,"%u",le_eep(END_TC_MAX_FCTR));
      env_lcd_d(223);//envia  simbolo de graus
      printf(lcd_putc,"C ");//envia C

      if(!sw_inc){//testa sw dec
      dbc_sw_inc();//aguarda soltar boto
      val=le_eep(END_TC_MAX_FCTR);
      if(val<LIMITE_SUP_TC_MAX){
       val++;
       escreve_eep(END_TC_MAX_FCTR,val);
      }
     }
     if(!sw_dec){//testa ws dec
      dbc_sw_dec();//aguarda soltar boto
      val=le_eep(END_TC_MAX_FCTR);
      if(val>LIMITE_INF_TC_MAX){
       val--;
       escreve_eep(END_TC_MAX_FCTR,val);
      }
     }

       if(!sw_setup){passo++; dbc_sw_setup();}
       delay_dbc();
     }//if(passo==6)

      if(passo==7) {env_lcd_c(1); passo++;}//limpa tela

//=========================================================================
  //     Ajusta Temperatura Mnima do Coletor Solar p/ ligar Bomba
  //     Obs. Evita congelamento da gua no coletor em dias muito frios
  //     fazendo a gua circular.
  //=========================================================================

    if(passo==8) {
      printf(lcd_putc,"\rNew TcMin Setup:");
      env_lcd_c(0xc6);//posiciona print
      printf(lcd_putc,"%u",le_eep(END_TC_MIN_FCTR));
      env_lcd_d(223);//envia  simbolo de graus
      printf(lcd_putc,"C ");//envia C

      if(!sw_inc){//testa sw dec
      dbc_sw_inc();//aguarda soltar boto
      val=le_eep(END_TC_MIN_FCTR);
      if(val<LIMITE_SUP_TC_MIN){
       val++;
       escreve_eep(END_TC_MIN_FCTR,val);
      }
     }
     if(!sw_dec){//testa ws dec
      dbc_sw_dec();//aguarda soltar boto
      val=le_eep(END_TC_MIN_FCTR);
      if(val>LIMITE_INF_TC_MIN){
       val--;
       escreve_eep(END_TC_MIN_FCTR,val);
      }
     }

       if(!sw_setup){passo++; dbc_sw_setup();}
       delay_dbc();
     }//if(passo==8)
 //=========================================================================
 //            Finaliza Setup e volta a rotina principal
 //=========================================================================
      if(passo==9) {
        env_lcd_c(1);//limpa tela
        printf(lcd_putc,"\rSettings saved! ");
        delay_ms(1000);//mostra mensagem por 1 segundo
        env_lcd_c(1);//limpa tela
        }

   }//while(passo<9)
}

/*******************************************************************************
              Rotina Main com Configuraes de Portas/Registradores
*******************************************************************************/

void main() {

   setup_adc_ports(sAN3 |sAN7 | VSS_VDD);  //seleciona AN2 e 5v como referncia
   setup_adc(ADC_CLOCK_DIV_8 );// ajusta o clock do adc
   setup_counters(RTCC_INTERNAL,RTCC_DIV_1);
   setup_timer_1(T1_DISABLED);
   setup_vref(FALSE);
   trisc=0b00111000;
   portc=0;
   trisa=0b011000;
   porta=0;
   cmcon0=7;
   cmcon1=0;
   lcd_init();//inicializa o lcd
   printf(lcd_putc,"\rTermostato_Difer");
   printf(lcd_putc,"\nLarios.Tecnol.ws");
   prog_special_char();//programa caracteres especiais na CGRAM
   delay_ms(2000);
   env_lcd_c(1);//apaga lcd
   f_ligado=0;
   f_gelo=0;
   f_sup_aq=0;
/*******************************************************************************
                 Lao Principal de Repetio
*******************************************************************************/

   for(;;){
     unsigned int8 a;
     testa_modo_prog();
     Tcol=get_temp(3);
     printf(lcd_putc,"\rTc:%03.1f",Tcol); //imprime no lcd
     env_lcd_d(223);//envia  (caracter especial)
     printf(lcd_putc,"C ");//envia C
     env_lcd_c(0x8b);//posiciona
     env_lcd_d(6); //seta
     env_lcd_d(':'); //2 pontos
     if(f_ligado){printf(lcd_putc," ON");}//envia ON
     else{printf(lcd_putc,"OFF ");}//envia Off

     T_reserv=get_temp(7);
     printf(lcd_putc,"\nTr:%03.1f",T_reserv); //imprime no lcd
     env_lcd_d(223);//envia  (caracter especial)
     printf(lcd_putc,"C ");//envia C
     env_lcd_c(0xcb);//posiciona
     env_lcd_d(CGRAM_QUINA_DOWNLEFT);
     env_lcd_d(CGRAM_TRACO_A);
     env_lcd_d(CGRAM_TRACO_A);
     env_lcd_d(CGRAM_TRACO_A);
     env_lcd_d(CGRAM_QUINA_UPRIGHT);

     for (a=0;a<10;a++){testa_modo_prog();delay_dbc();}

     if(f_ligado){
     env_lcd_c(0xcb);//posiciona
     env_lcd_d(CGRAM_QUINA_UPLEFT);
     env_lcd_d(CGRAM_TRACO_B);
     env_lcd_d(CGRAM_TRACO_B);
     env_lcd_d(CGRAM_TRACO_B);
     env_lcd_d(CGRAM_QUINA_DOWNRIGHT);
     }
     for (a=0;a<10;a++){testa_modo_prog();delay_dbc();}
     dt_on=le_eep(END_DELTA_ON);//delta on
     dt_off=le_eep(END_DELTA_OFF);//delta off
     sup_aq=le_eep(END_TC_MAX_FCTR);//Superaquecimento
     gelo=le_eep(END_TC_MIN_FCTR);//congelamento

     if(Tcol>=(float)sup_aq){f_sup_aq=1;}//superaquecimento no coletor solar
     else{
     if(Tcol<=(float)(sup_aq-dt_off))f_sup_aq=0;//sem superaquecimento
     }

     if(Tcol<=(float)gelo)f_gelo=1;//baixa temperatura no coletor solar
     if(Tcol>=(float)(dt_off+gelo))f_gelo=0;//sem formao de gelo

     //testa se deve setar flag de ligado
      if(!f_sup_aq){
         if((Tcol>=(T_reserv+ (float)dt_on))) f_ligado=1;
      }
      else{ f_ligado=0;}//desliga bomba por superaquecimento


      //testa por formao de gelo no coletor solar
       if(!f_gelo){
       if(Tcol<=(T_reserv+(float)dt_off))f_ligado=0;
       }
       else{ f_ligado=1;}

     //aciona sada da bomba
     if(f_ligado)rele=1;else rele=0;

   }//for(;;)
}//void main()
